import {
	Transform,
	Type,
	instanceToPlain, plainToInstance
} from 'class-transformer';

function stringCompare(a: string, b: string) {
	return a.localeCompare(b, undefined, {
		sensitivity: 'accent'
	}) === 0;
}

interface IBingoGame {
	id: string;
	name: string;
	description: string;
	generator: 'simple' | 'srl_v5' | 'srl_v8';
	goals: IBingoGoal[];
	groups: IBingoGroup[];
}

interface IBingoGoal {
	name: string;
	group?: IBingoGroup;
	tags: string[];
	possible_spaces: number[];
}

interface IBingoGroup {
	name: string;
	parent?: IBingoGroup;
}

class BingoGame implements IBingoGame {
	id: string;

	name: string;

	description: string;

	generator: 'simple' | 'srl_v5' | 'srl_v8' = 'simple';

	@Type(() => BingoGoal)
		goals: IBingoGoal[] = [];

	@Type(() => BingoGroup)
		groups: IBingoGroup[] = [];

	constructor(id: string, name: string, description: string) {
		this.id = id;
		this.name = name;
		this.description = description;
	}

	addGoal(goal: IBingoGoal): void {
		if (this.goals.some(_goal => stringCompare(_goal.name, goal.name)))
			return;

		this.goals.push(goal);
	}

	removeGoal(goal: IBingoGoal): void {
		this.goals = this.goals.filter(_goal => !stringCompare(_goal.name, goal.name));
	}

	addGroup(group: IBingoGroup): void {
		if (this.groups.some(_group => stringCompare(_group.name, group.name)))
			return;

		this.groups.push(group);
	}

	removeGroup(group: IBingoGroup): boolean {
		// Group has parent, don't delete
		if (this.groups.some(_group => _group.parent && stringCompare(_group.parent.name, group.name)))
			return false;

		if (this.goals.some(goal => goal.group && stringCompare(goal.group.name, group.name)))
			return false;
		// TODO: Make sure no goal is still using group;
		// boolean override "force_remove" which will remove the group from all goals
		console.log('deltee', group);
		this.groups = this.groups.filter(_group => !stringCompare(_group.name, group.name));
		console.log(this.groups);

		return true;
	}
}

class BingoGoal implements IBingoGoal {
	// id: string; // uuid-by-string maybe? https://www.npmjs.com/package/uuid-by-string

	name: string;

	group?: IBingoGroup;

	tags: string[] = [];

	possible_spaces: number[] = [];

	constructor(name: string, group?: BingoGroup) {
		this.name = name;
		this.group = group;
	}

	addTag(tag: string): IBingoGoal {
		if (this.tags.includes(tag))
			return this;

		this.tags.push(tag);

		return this;
	}

	removeTag(tag: string): void {
		this.tags = this.tags.filter(_tag => !stringCompare(_tag, tag));
	}

	setGroup(group: BingoGroup) {
		this.group = group;
	}

	setPossibleSpaces(possible_spaces: number[]) {
		this.possible_spaces = possible_spaces;
	}
}

class BingoGroup implements IBingoGroup {
	name: string;

	@Type(() => BingoGroup)
		parent?: IBingoGroup;

	constructor(name: string, parent?: BingoGroup) {
		this.name = name;
		this.parent = parent;
	}
}

const game = new BingoGame('yakuza_0', 'Yakuza 0', 'The funny game we used for all bingos.');
const battle_group = new BingoGroup('Battle');
const defeat_enemies = new BingoGroup('Defeat enemies on the street', battle_group);
game.addGroup(battle_group);
game.addGroup(defeat_enemies);

const easy = new BingoGoal('Defeat 50 enemies on the street', defeat_enemies).addTag('Easy');
const normal = new BingoGoal('Defeat 150 enemies on the street', defeat_enemies).addTag('Normal');
const hard = new BingoGoal('Defeat 300 enemies on the street', defeat_enemies).addTag('Hard');

game.addGoal(easy);
game.addGoal(normal);
game.addGoal(hard);

const jsonifiedInstance = instanceToPlain(game);
// console.log(JSON.stringify(jsonifiedInstance));
// console.log(JSON.stringify(game));

const json_string = '{"id":"yakuza_0","name":"Yakuza 0","description":"The funny game we used for all bingos.","generator":"simple","goals":[{"name":"Defeat 50 enemies on the street","tags":["Easy"],"group":{"name":"Defeat enemies on the street","parent":{"name":"Battle"}},"possible_spaces":[]},{"name":"Defeat 150 enemies on the street","tags":["Normal"],"group":{"name":"Defeat enemies on the street","parent":{"name":"Battle"}},"possible_spaces":[]},{"name":"Defeat 300 enemies on the street","tags":["Hard"],"group":{"name":"Defeat enemies on the street","parent":{"name":"Battle"}},"possible_spaces":[]}],"groups":[{"name":"Battle"},{"name":"Defeat enemies on the street","parent":{"name":"Battle"}}]}';
const parsed: JSON = JSON.parse(json_string);
const parsed_game = plainToInstance(BingoGame, parsed);
// console.log(JSON.stringify(other_game) === JSON.stringify(jsonifiedInstance));

parsed_game.removeGroup(defeat_enemies);
console.log(parsed_game.goals);
